/*
 * Copyright (C) 2006-2007	Andre Beckedorf
 * 					 		<evilJazz _AT_ katastrophos _DOT_ net>
 *
 * This file is based on eMotion by Michael Swieton <mike@swieton.net>.
 * The few remaining parts in this almost completely rewritten file are
 * Copyright (C) 2005 Atmark <atmarkat _AT_ msn _DOT_ com>
 *                    AGAWA Koji <i _AT_ atty _DOT_ jp>
 *                    Michael Swieton <mike@swieton.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <qapplication.h>
#include <qfileinfo.h>
#include <qstring.h>
#include <qfile.h>

#include <cstdio>
#include <cstdlib>

#include <signal.h>
#include <unistd.h>
#include <errno.h>

#ifdef WINDOWS
#include "wincompat.h"
#endif

#include "mplayerbackend.h"
#include "helpers.h"
#include "media.h"
#include "configuration.h"

//#define DEBUG

#include "debug.h"
#include "compathack.h"

static QProcess *checkFeaturesProcess;

// available before version 1.0pre8:
static bool mplayerHasGetPercentPos;
static bool mplayerHasGetTimeLength;

// added in version 1.0pre8:
static bool mplayerHasGetTimePos;
static bool mplayerCanQueueFiles;
static bool mplayerSupportsIdle;
static bool mplayerSupportsPausingKeepToogle;

MPlayerBackend::MPlayerBackend(MPlayerConfiguration &config)
    :	AbstractPlaybackBackend(),
        config_(config),
        quitState_(QuitNull),
        process_(NULL),
        lastMplayerBinLocation_(""),
        buffer_(""),
        messages_(""),
        params_(""),
        isPaused_(false),
        playbackWasStarted_(false),
        errorDetected_(false),
        isSendingPTS_(false),
        parsing_(false),
        watchdogCheck_(0),
        newPlaytime_(-1)
{
    DGUARDMETHODTIMED;
    register_handler(SIGTERM, this);
    register_handler(SIGSEGV, this);

    connect(this, SIGNAL(shutdown()), this, SLOT(cleanUp()));
    connect(this, SIGNAL(killed()), this, SLOT(cleanUp()));

    connect(&timer_, SIGNAL(timeout()), this, SLOT(pollMPlayer()));

    connect(&SystemVolume::singleton(), SIGNAL(levelChanged(int)), this, SLOT(systemVolumeChanged(int)));

    connect(&killTimer_, SIGNAL(timeout()), this, SLOT(killWaitingProcesses()));

#ifndef QT4
    processKillWaitList_.setAutoDelete(true);
#endif

    osdTimer_ = new QTimer(this);
    connect(osdTimer_, SIGNAL(timeout()), this, SLOT(osdTimerTimedUpdate()));
}

MPlayerBackend::~MPlayerBackend()
{
    DGUARDMETHODTIMED;
    quitMplayer(false);

#ifdef QT4
    qDeleteAll(processKillWaitList_);
#endif

    unregister_handler(this);
}

void MPlayerBackend::checkFeaturesRunProcess(const QString &program, const QStringList &args)
{
    // timeout timer...
    QTime time;
    time.start();
    bool timedOut = false;

#ifdef QT4
    checkFeaturesProcess->setProcessChannelMode(QProcess::MergedChannels);
    checkFeaturesProcess->setReadChannel(QProcess::StandardOutput);
    checkFeaturesProcess->start(program, args);
    if (!checkFeaturesProcess->waitForStarted(-1))
    {
        DPRINTF("Could not launch mplayer!");
        return;
    }
#else
    QStringList arguments(args);
    arguments.insert(arguments.begin(), program);
    checkFeaturesProcess->setArguments(arguments);
    checkFeaturesProcess->start();
#endif

#ifdef QT4
    while ((checkFeaturesProcess->state() != QProcess::NotRunning || checkFeaturesProcess->canReadLine()) && !timedOut)
#else
    while ((checkFeaturesProcess->isRunning() || checkFeaturesProcess->canReadLineStdout()) && !timedOut)
#endif
    {
        DPRINTF("Feature check loop, trying to read lines...");
        timedOut = time.elapsed() > 5000;
        checkFeaturesProcessParseLines();
#ifdef QT4
        checkFeaturesProcess->waitForReadyRead(1);
#else
        usleep(1000); // sleep 1 ms
#endif
    }

#ifdef QT4
    if (checkFeaturesProcess->state() != QProcess::NotRunning)
    {
        checkFeaturesProcess->terminate();
        checkFeaturesProcess->waitForFinished(100);
    }

    if (checkFeaturesProcess->state() != QProcess::NotRunning)
    {
        checkFeaturesProcess->kill();
        checkFeaturesProcess->waitForFinished(100);
    }
#else
    if (checkFeaturesProcess->isRunning())
    {
        checkFeaturesProcess->tryTerminate();
        usleep(50000); // wait 50 ms
    }

    if (checkFeaturesProcess->isRunning())
    {
        checkFeaturesProcess->kill();
        usleep(100000); // wait 100 ms
    }
#endif
}

bool MPlayerBackend::checkFeatures()
{
    DGUARDMETHODTIMED;

    QString mplayerBinLocation = config_.mplayerBinLocation();

    if (lastMplayerBinLocation_ != mplayerBinLocation)
    {
        mplayerHasGetTimePos = false;
        mplayerHasGetTimeLength = false;
        mplayerHasGetPercentPos = false;
        mplayerCanQueueFiles = false;
        mplayerSupportsIdle = false;
        mplayerSupportsPausingKeepToogle = false;

        QStringList args;
        checkFeaturesProcess = new QProcess();

        args << "-input" << "cmdlist";
        checkFeaturesRunProcess(mplayerBinLocation, args);

        args.clear();
        args << "-list-options";
        checkFeaturesRunProcess(mplayerBinLocation, args);

        delete checkFeaturesProcess;
        checkFeaturesProcess = NULL;

        if (mplayerSupportsIdle)
        {
            mplayerSupportsPausingKeepToogle = true;
        }

        lastMplayerBinLocation_ = mplayerBinLocation;

        return true;
    }

    return false;
}

void MPlayerBackend::checkFeaturesProcessParseLines()
{
#ifdef QT4
    while (checkFeaturesProcess->canReadLine())
    {
        QString line = checkFeaturesProcess->readLine();
#else
    while (checkFeaturesProcess->canReadLineStdout())
    {
        QString line = checkFeaturesProcess->readLineStdout();
#endif
        DPRINTF("LINE stdout: %s", (const char*)(line.toUtf8()));

        if (line.startsWith("get_time_pos"))
        {
            mplayerHasGetTimePos = true;
            qDebug("mplayer supports get_time_pos.");
        }
        else if (line.startsWith("get_time_length"))
        {
            mplayerHasGetTimeLength = true;
            qDebug("mplayer supports get_time_length.");
        }
        else if (line.startsWith("get_percent_pos"))
        {
            mplayerHasGetPercentPos = true;
            qDebug("mplayer supports get_percent_pos.");
        }
        else if (line.startsWith("loadfile") && line.indexOf("[Integer]") != -1)
        {
            mplayerCanQueueFiles = true;
            qDebug("mplayer supports advanced syntax for loadfile, i.e. it can queue files.");
        }
        else if (line.simplified().startsWith("idle Flag"))
        {
            mplayerSupportsIdle = true;
            qDebug("mplayer supports idling in slave mode.");
        }
    }
}

QStringList MPlayerBackend::getArgListForMedia(Media *media)
{
    QStringList args;

    args << config_.mplayerBinLocation();
    args << "-slave";

    if (mplayerSupportsIdle)
    {
        args << "-idle";
        args << "-v";
    }

    args << "-quiet";
#ifdef QTOPIA
    args << "-nortc";
#endif

    if (media->isVideo())
    {
        // Video Output
        if (config_.isOverlayEnabled)
        {
            switch (config_.machineCategory())
            {
            case MPlayerConfiguration::MACHINE_SLC700:
                args << "-vo";
                args << "w100";
                break;
            case MPlayerConfiguration::MACHINE_SLC3000:
                args << "-vo";
                args << "bvdd";
                break;
            case MPlayerConfiguration::MACHINE_GENERIC:
                break;
            }
        }

        // Full Screen
        if (config_.isFullScreenEnabled) {
            switch (config_.machineCategory()) {
            case MPlayerConfiguration::MACHINE_SLC700:
                args << "-fs";
                break;
            case MPlayerConfiguration::MACHINE_SLC3000:
                args << "-vm";
                break;
            case MPlayerConfiguration::MACHINE_GENERIC:
                args << "-fs";
                break;
            }
        }

        // Double Buffering
        if (!config_.isDoubleBufferingEnabled)
            args << "-nodouble";
        else
            args << "-double";

        // Frame Drop
        switch (config_.framedropMethod)
        {
        case MPlayerConfiguration::FRAMEDROP_NORMAL:
            args << "-framedrop";
            break;
        case MPlayerConfiguration::FRAMEDROP_HARD:
            args << "-hardframedrop";
            break;
        default:
            break;
        }

        if (useSoftVolume())
        {
            args << "-softvol";
            //args << "-volume";
            //args << QString::number(qSystemVolume.getCurrentLevel());
        }

#ifdef QTOPIA
        // 映像と音声の同期を鈍感にする。
        // 一時的に負荷が高まる映像の再生が安定する。
        args << "-autosync";
        args << "100";

        // FIXME: more generic handling
        if (media->width() == 352 && media->height() == 240)
        {
            args << "-vf-add";
            args << "crop=320:240";
        }

        if (media->width() > media->height())
        {
            if (media->width() > qApp->desktop()->width() ||
                media->height() > qApp->desktop()->height())
            {
                args << "-vf-add";
                args << "crop=640:480";	// FIXME: use qApp->desktop()
            }
        }

        if (!config_.isOverlayEnabled)
        {
            if (media->width() > media->height())
            {
                args << "-vf-add";
                args << "rotate=1";
            }
        }
#endif
    }

#ifdef QTOPIA
    // これらのサンプリングレート以外をそのまま再生しようとすると非常に
    // 負荷が高くなるので、44100KHzにリサンプリングする
    if (media->audioRate() != 44100 && media->audioRate() != 48000 && media->audioRate() != 22050)
    {
        args << "-af";
        args << "lavcresample=44100";
    }
#endif

    // Cache handling...
    QString cacheSize;
    if (config_.isCacheEnabled)
    {
        args << "-cache";
        cacheSize = QString::number(config_.cacheSize);
        args << cacheSize;

        if (media->location().isLocalFile())
        {
            // for local files, start playback immediately, ie.
            // don't wait till the cache is filled up.
            // We assume local files are read faster than they are played back.
            args << "-cache-min";
            args << "0";
        }
    }
    else
        args << "-nocache";

    // 初期状態ではOSDを表示しない
    args << "-osdlevel";
    args << "0";

    QString fileExtension = media->location().fileName();
    fileExtension = fileExtension.right(fileExtension.length() -1 - fileExtension.lastIndexOf('.'));
    DPRINTF("fileExtension: %s", (const char*)fileExtension.toLatin1());
    QString customArguments = config_.mplayerArguments(fileExtension);
    if (!customArguments.isEmpty())
    {
        DPRINTF("customArguments: %s", (const char*)customArguments.toLatin1());
#ifdef QT4
        QStringList opts(customArguments.split(' ', QString::SkipEmptyParts));
#else
        QStringList opts(QStringList::split(' ', customArguments));
#endif

        for (QStringList::Iterator it = opts.begin(); it != opts.end(); ++it)
            args << (*it);
    }

    // その他の追加オプション
    if (config_.isAdditionalOptionsEnabled)
    {
#ifdef QT4
        QStringList opts(config_.additionalOptions.split(' ', QString::SkipEmptyParts));
#else
        QStringList opts(QStringList::split(' ', config_.additionalOptions));
#endif
        for (QStringList::Iterator it = opts.begin(); it != opts.end(); ++it)
            args << (*it);
    }

    return args;
}

bool MPlayerBackend::canRecycleMplayer(Media *media)
{
    QStringList args;
    return canRecycleMplayer(media, args);
}

bool MPlayerBackend::canRecycleMplayer(Media *media, QStringList &args)
{
    args = getArgListForMedia(media);

    bool mplayerBinLocationUpdated = checkFeatures();

    bool argsDiffer = args.count() != currentProcessArgs_.count();
    if (!argsDiffer)
        for (int i = 0; i < args.count(); ++i)
            if (args[i] != currentProcessArgs_[i])
            {
                argsDiffer = true;
                DPRINTF("%s != %s", (const char *)(currentProcessArgs_[i].toLatin1()), (const char *)(args[i].toLatin1()));
                break;
            }

    //return false; // for manual override during debugging...
    return isRunning() && !argsDiffer && !media->isVideo() && !mplayerBinLocationUpdated;
}

void MPlayerBackend::killWaitingProcesses()
{
    DGUARDMETHOD;
    DPRINTF("It's killing season: %d", processKillWaitList_.count());
    processKillWaitList_.clear();
    DPRINTF("All dead: %d", processKillWaitList_.count());
}

void MPlayerBackend::scheduleDestructionOfProcess(QProcess *process)
{
#ifdef QT4
    disconnect(process, SIGNAL(finished(int, QProcess::ExitStatus)));
    disconnect(process, SIGNAL(readyRead()));

    if (!processKillWaitList_.contains(process))
#else
    disconnect(process, SIGNAL(processExited()));
    disconnect(process, SIGNAL(readyReadStdout()));

    if (!processKillWaitList_.containsRef(process))
#endif
    {
        processKillWaitList_.append(process);
#ifdef QT4
        killTimer_.setSingleShot(true);
        killTimer_.start(100);
#else
        killTimer_.start(100, true);
#endif
    }
}

AbstractPlaybackBackend::PlayResult MPlayerBackend::play(Media *media)
{
    DGUARDMETHODFTIMED("[%s]", (const char *)media->location().toString().toUtf8());

#ifdef QT4
    //QASSERT(media != 0);
#else
    ASSERT(media != 0);
#endif

    /*
    if (media->location().isLocalFile() && !QFile::exists(media->location().toString()))
    {
        DPRINTF("%s was not found", QString(media->location()).toLatin1());
        return AbstractPlaybackBackend::MediaNotFound;
    }
    */

    if (!media->isSupported())
    {
        DPRINTF("%s is Not supported", (const char *)media->location().toString().toLatin1());
        return AbstractPlaybackBackend::MediaNotSupported;
    }

    DPRINTF("Determining mplayer binary features...");
    checkFeatures();

    QStringList args;

    if (!canRecycleMplayer(media, args))
    {
        if (isRunning())
            quitMplayer(false);

        // currentProcessArgs may only contain command line
        // arguments without the actual filename of the media
        // to play. This is mandatory because we use it to compare
        // against it next time we try to recycle the MPlayerBackend process.
        currentProcessArgs_ = args;

        args << media->location().toString();

        params_ = args.join(" ") + "\n";
        messages_ = QString::null;
        playbackWasStarted_ = false;
        errorDetected_ = false;
        quitState_ = QuitNull;

        if (qConfig.isBackendLoggingEnabled)
        {
            logStream_.setDevice(new QFile(qConfig.backendLogFileName));
#ifdef QT4
            logStream_.device()->open(QIODevice::WriteOnly|QIODevice::Truncate);
#else
            logStream_.device()->open(IO_WriteOnly|IO_Truncate);
#endif
            logStream_ << params_;
#ifndef QT4
            logStream_.device()->flush();
#endif
        }

        isSendingPTS_ = false;

        DPRINTF("Starting mplayer slave...");
#ifdef QT4
        process_ = new QProcess();
        process_->setProcessChannelMode(QProcess::MergedChannels);
        process_->setReadChannel(QProcess::StandardOutput);
        connect(process_, SIGNAL(readyRead()), this, SLOT(processReadyReadStdout()));
        connect(process_, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(processExited()));

        QString program = args.first();
        args.removeFirst();

        process_->start(program, args);
        if (!process_->waitForStarted(-1))
        {
#else
        process_ = new QProcess(args);
        process_->setCommunication(QProcess::Stdin|QProcess::Stdout|QProcess::Stderr|QProcess::DupStderr);
        connect(process_, SIGNAL(readyReadStdout()), this, SLOT(processReadyReadStdout()));
        connect(process_, SIGNAL(processExited()), this, SLOT(processExited()));

        if (!process_->start())
        {
#endif
            DPRINTF("Could not execute mplayer");
            return AbstractPlaybackBackend::InitializationFailed;
        }
    }
    else
    {
        params_ = "Using recycled mplayer process.\n" + args.join(" ") + "\n";
        messages_ = QString::null;
        playbackWasStarted_ = false;
        errorDetected_ = false;
        quitState_ = QuitNull;

        DPRINTF("Recycling mplayer slave");
        QString command = "loadfile \"" + media->location().toString() + "\"";

        if (mplayerSupportsPausingKeepToogle && isPaused_)
            command = "pausing_keep " + command + "\npause";

        sendCommand(command);
        isPaused_ = false;
    }

    sendCurrentVolume();

    watchdogCheck_ = 0;
    timer_.start(1000);

    emit started();

    isOSDVisible_ = false;
    isTemporaryOSDEnabled_ = false;

    return AbstractPlaybackBackend::PlaybackInitiated;
}

void MPlayerBackend::processReadyReadStdout()
{
    DGUARDMETHOD;
    if (process_)
        parseLines();
}

void MPlayerBackend::processExited()
{
    DGUARDMETHOD;

    if (process_)
    {
        parseLines();

        currentProcessArgs_.clear();

        DPRINTF("emit errorMessage?");

        if (!messages_.isNull() && !playbackWasStarted_ && (process_->exitStatus() != 0 || errorDetected_))
        {
            DPRINTF("emit errorMessage!");
            emit errorMessage(params_ + "\n" + messages_);
        }
    }
}

void MPlayerBackend::sendCommand(const QString &c)
{
    DENTERMETHODF("[%s]", (const char *)c.toUtf8());

    if (!isRunning())
    {
        DEXITMETHODF("[%s] MPlayer not running!", (const char *)c.toUtf8());
        return;
    }

    writeToProcess(process_, c);

    DEXITMETHODF("[%s]", (const char *)c.toUtf8());
}

void MPlayerBackend::stop(bool emitQuitSignal)
{
    DENTERMETHODF("[%s]", emitQuitSignal ? "true" : "false");

    if (!isRunning())
    {
        DEXITMETHODF("[%s] MPlayer not running!", emitQuitSignal ? "true" : "false");
        return;
    }

    if (mplayerSupportsIdle)
    {
        sendCommand("stop");
        timer_.stop();

        // if playback is paused and we are stopping,
        // keep the isPaused flag, so we can handle it correctly in the play method
        if (!mplayerSupportsPausingKeepToogle)
            isPaused_ = false;

        buffer_ = "";
        playbackWasStarted_ = false;
        errorDetected_ = false;
    }
    else
        quitMplayer(emitQuitSignal);

    DEXITMETHODF("[%s]", emitQuitSignal ? "true" : "false");
}

void MPlayerBackend::quitMplayer(bool emitQuitSignal)
{
    DENTERMETHODF("[%s]", emitQuitSignal ? "true" : "false");

    if (!isRunning())
    {
        DEXITMETHODF("[%s] MPlayer not running!", emitQuitSignal ? "true" : "false");
        return;
    }

    if (!emitQuitSignal)
        blockSignals(true);

    // timeout timer...
    QTime time;
    time.start();
    bool timedOut = false;

    sendCommand("quit");
    while (isRunning() && !timedOut)
    {
        timedOut = time.elapsed() > 500;
        usleep(10000); // sleep 10 ms
    }

    if (timedOut)
    {
        timedOut = false;
        // try to terminate the process if it hasn't
        // already exited...
#ifdef QT4
        process_->terminate();
#else
        process_->tryTerminate();
#endif

        // wait for shutdown, give process time to
        // clean up.
        while (isRunning() && !timedOut)
        {
            timedOut = time.elapsed() > 500;
            usleep(10000); // sleep 10 ms
        }

        // if friendly termination didn't work out,
        // kill the process.
        if (timedOut)
        {
            process_->kill(); // final judgement .
            usleep(10000); // sleep 10 ms
        }
    }

    process_->kill(); // final judgement .
    currentProcessArgs_.clear();

    if (process_)
    {
        scheduleDestructionOfProcess(process_);
        process_ = NULL;
    }

    emit shutdown(); // also triggers cleanUp();

    if (!emitQuitSignal)
        blockSignals(false);

    DEXITMETHODF("[%s]", emitQuitSignal ? "true" : "false");
}

void MPlayerBackend::cleanUp()
{
    DENTERMETHOD;

    if (logStream_.device())
    {
        delete logStream_.device();
#ifdef QT4
        logStream_.setDevice(0);
#else
        logStream_.unsetDevice();
#endif
    }

    timer_.stop();
    isPaused_ = false;
    buffer_ = "";
    currentProcessArgs_.clear();

    DEXITMETHOD;
}

void MPlayerBackend::parseLines()
{
    if (parsing_)
        return;

    DENTERMETHOD;

    int newPlaytime = -1;

#ifdef QT4
    while (process_ && process_->canReadLine())
    {
        parsing_ = true;
        QString line = process_->readLine();
#else
    while (process_ && process_->canReadLineStdout())
    {
        parsing_ = true;
        QString line = process_->readLineStdout();
#endif
        DPRINTF("LINE stdout: %s", (const char*)(line.toUtf8()));

        if (!playbackWasStarted_)
            messages_ += line + "\n";

        if (logStream_.device())
        {
            logStream_ << line << "\n";
#ifndef QT4
            logStream_.device()->flush();
#endif
        }

        if (line.startsWith("PTS=") && !mplayerHasGetTimePos)
        {
            if (!isSendingPTS_)
            {
                DPRINTF("mplayer is sending PTS.");
                isSendingPTS_ = true;
            }

            int playtime = line.mid(4).toInt();

            playbackWasStarted_ = true;

            newPlaytime = playtime;
        }
        else if (line.startsWith("ANS_TIME_POSITION="))
        {
            QString positionString = line.mid(18);
            int posDot = positionString.indexOf(".");
            int playtime = 0;

            if (posDot > -1)
            {
                playtime = positionString.left(posDot).toInt();
                if (positionString.mid(posDot + 1, 1).toInt() > 5)
                    ++playtime;
            }
            else
                playtime = positionString.toInt();

            playbackWasStarted_ = true;
            watchdogCheck_ = 0;

            newPlaytime = playtime;
        }
        else if (mplayerSupportsIdle && line.indexOf("EOF code: 1") != -1)
        {
            quitState_ = QuitEOF;
            QTimer::singleShot(1, this, SLOT(pollMPlayer()));
            newPlaytime = -1;
        }
        else if (line.startsWith("Exiting..."))
        {
            DPRINTF("OMG, I read Exiting!");

            // ここですぐにemitしてもいい気がする
            if (line.indexOf("Quit") != -1)
            {
                quitState_ = QuitUser;
            }
            else if (line.indexOf("End of file") != -1)
            {
                quitState_ = QuitEOF;
            }

            QTimer::singleShot(1, this, SLOT(pollMPlayer()));

            newPlaytime = -1;
        }
        else if (line.startsWith("Unknown option") ||
                line.startsWith("FATAL") ||
                line.startsWith("Cannot find codec") ||
                line.startsWith("File not found ") ||
                line.startsWith("Failed to open ") ||
                line.startsWith("- MPlayer crashed") ||
                line.startsWith("MPlayer interrupted by signal"))
        {
            playbackWasStarted_ = false;
            errorDetected_ = true;
            newPlaytime = -1;
        }
        else if (line.startsWith("Seek failed")) {
            // -hr-mp3-seekのとき、末尾以降にシークしようとするとSeek failedの
            // メッセージを吐きながら帰ってこなくなる。mplayer側の修正が必要。
            stop();
        }

        emit output(line);
    }

    parsing_ = false;

    if (newPlaytime > -1)
    {
        newPlaytime_ = newPlaytime;
        QTimer::singleShot(1, this, SLOT(updatedValues()));
    }

    DEXITMETHOD;
}

void MPlayerBackend::updatedValues()
{
    if (newPlaytime_ > -1)
    {
        DPRINTF("emit currentPlayTime(%d)", newPlaytime_);
        emit currentPlayTime(newPlaytime_);
        newPlaytime_ = -1;
    }
}

void MPlayerBackend::pause()
{
    if (!isRunning())
        return;

    enableTemporaryOSD();
    sendCommand("pause");

    if (isPaused_) {
        isPaused_ = false;
        emit resumed();
    } else {
        isPaused_ = true;
        emit paused();
    }
}

void MPlayerBackend::seekRelativeSeconds(int seconds)
{
    QString command;
    enableTemporaryOSD();
    sendCommand(command.sprintf("seek %d 0", seconds));
    if (isPaused_)
        sendCommand("pause");

    updatePosition();
}

void MPlayerBackend::seekAbsolutePercentage(int percent)
{
    QString command;
    sendCommand(command.sprintf("seek %d 1", percent));
    if (isPaused_)
        sendCommand("pause");

    updatePosition();
}

void MPlayerBackend::seekAbsoluteSeconds(int seconds)
{
    QString command;
    sendCommand(command.sprintf("seek %d 2", seconds));
    if (isPaused_)
        sendCommand("pause");

    updatePosition();
}

void MPlayerBackend::osdShowProgressBar(int seconds)
{
    QString command;
    sendCommand(command.sprintf("osd_show_progbar %d", seconds));
}

void MPlayerBackend::toggleOSD()
{
    if (isTemporaryOSDEnabled_)
        return;

    if (isOSDVisible_)
    {
        isOSDVisible_ = false;
        sendCommand("osd 0");
    }
    else
    {
        isOSDVisible_ = true;
        sendCommand("osd 3");
    }
}

void MPlayerBackend::enableTemporaryOSD()
{
    if (isTemporaryOSDEnabled_)
        return;

    savedOSDVisiblity_ = isOSDVisible_;
    if (!isOSDVisible_)
        toggleOSD();

    isTemporaryOSDEnabled_ = true;

#ifdef QT4
    osdTimer_->setSingleShot(true);
    osdTimer_->start(3000);
#else
    osdTimer_->start(3000, true);
#endif
}

void MPlayerBackend::disableTemporaryOSD()
{
    isTemporaryOSDEnabled_ = false;

    if (savedOSDVisiblity_ != isOSDVisible_)
        toggleOSD();
}

void MPlayerBackend::osdTimerTimedUpdate()
{
    if (!isPaused())
        disableTemporaryOSD();
}

void MPlayerBackend::setBrightness(int value)
{
    QString args;
    sendCommand(args.sprintf("brightness %d", value));
}

void MPlayerBackend::pollMPlayer()
{
    DENTERMETHOD;

    if (!isRunning())
    {
        if (quitState_ == QuitEOF)
        {
            DPRINTF("QuitEOF");
            cleanUp();
            emit eof();
        }
        else
        {
            DPRINTF("QuitUser");
            emit shutdown(); // also triggers cleanUp();
        }
    }
    else if (mplayerSupportsIdle)
    {
        if (quitState_ == QuitEOF)
        {
            DPRINTF("EOF");
            quitState_ = QuitNull;
            timer_.stop();
            isPaused_ = false;
            buffer_ = "";
            playbackWasStarted_ = false;
            errorDetected_ = false;

            emit eof();
        }
        else
            updatePosition();
    }
    else
        updatePosition();

    DEXITMETHOD;
}

void MPlayerBackend::updatePosition()
{
    if (!isSendingPTS_ && !isPaused_)
    {
        if (mplayerHasGetTimePos)
        {
            sendCommand("get_time_pos");
            ++watchdogCheck_;
        }
        else if (mplayerHasGetPercentPos)
        {
            sendCommand("get_percent_pos");
            ++watchdogCheck_;
        }

        if (watchdogCheck_ > 5)
        {
            // For some reason mplayer doesn't send the current position
            DPRINTF("Something went wrong and MPlayer doesn't respond to our position requests, quitting MPlayer.");
            quitMplayer(true);
        }
    }
}

void MPlayerBackend::handleSignal(int sig)
{
    DGUARDMETHODF("[%d]", sig);
    DPRINTF("Got signal %d.", sig);
    quitMplayer(false);
    exit(-sig);
}

void MPlayerBackend::systemVolumeChanged(int level)
{
    sendCurrentVolume();
}

void MPlayerBackend::sendCurrentVolume()
{
    DGUARDMETHOD;
    if (isRunning() && useSoftVolume())
    {
        DPRINTF("Setting volume to %d...", qSystemVolume.getCurrentLevel());
        sendCommand("volume " + QString::number(qSystemVolume.getCurrentLevel()) + " 1");
    }
}
