/*
 * Copyright (C) 2010	Andre Beckedorf
 * 				 		<evilJazz _AT_ katastrophos _DOT_ net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef PLAYLIST_H_INCLUDED
#define PLAYLIST_H_INCLUDED

#include <qlist.h>
#include <qwidget.h>
#include <qpoint.h>

#include <QKeyEvent>
#include <QPixmap>
#include <QResizeEvent>
#include <QMouseEvent>
#include <QStyledItemDelegate>
#include <QTreeView>

#include "mediadatabase.h"
#include "media.h"
#include "playlistaspectlistviewrenderer.h"

class Media;
class PlayListView;
class PlaybackController;
class SkinModeInformation;

class PlayListItemDelegate: public QStyledItemDelegate
{
    Q_OBJECT
public:
    PlayListItemDelegate(PlayListView *playListView);

    void paint(QPainter *painter, const QStyleOptionViewItem &option, const QModelIndex &index) const;

private:
    friend class PlayListView;
    PlayListView *playListView_;
};

class PlayListView: public QTreeView
{
    Q_OBJECT
public:
    enum Columns {
        COLUMN_NONE = 0,
        COLUMN_TRACKNO = 1,
        COLUMN_TITLE = 2,
        COLUMN_ARTIST = 3,
        COLUMN_ALBUM = 4,
        COLUMN_LENGTH = 5,
        COLUMN_MAX = 6
    };

    enum InputMode {
        Select = 0,
        MultiSelect = 1,
        Move = 2
    };

    PlayListView(QWidget *parent = 0);
    virtual ~PlayListView();

    void attachTo(PlayListAspectListViewRenderer *source, PlaybackController *controller);
    void detach();
    PlayListAspectListViewRenderer *aspectSource() { return aspectSource_; }

    void beginUpdate();
    void endUpdate();

    PlayListView::InputMode inputMode() { return inputMode_; }
    void setInputMode(PlayListView::InputMode inputmode);

    int selectedItemsCount();
    MediaIDList getSelectedMediaIDs();
    PlayListIDList getSelectedItemIDs();
    MediaList getSelectedMedia();

    PlayListItem *currentPlayListItem();

    void makeItemVisibleByID(unsigned long playListID, bool selectItem = false);

    void insertMediaIDsAtCurrentIndex(const MediaIDList &srcMediaIDs);

    void setBackgroundColors(const QColor &evenColor, const QColor &oddColor);
    void setForegroundColor(const QColor &textColor);
    void setPlayIndicatorImage(const QPixmap &image);
    void setPauseIndicatorImage(const QPixmap &image);
    void setErrorIndicatorImage(const QPixmap &image);

    void setAutoScaleColumnsEnabled(bool value);
    bool autoScaleColumnsEnabled() { return autoScaleColumnsEnabled_; }

    void setPopupMenu(QMenu *popupMenu) { popupMenu_ = popupMenu; }
    QMenu *popupMenu() const { return popupMenu_; }

    void saveColumnOrder(const QString &configName, const QString &groupName);
    void saveColumnWidth(const QString &configName, const QString &groupName);
    void loadColumnOrder(const QString &configName, const QString &groupName);
    void loadColumnWidth(const QString &configName, const QString &groupName);

    void setSkinMode(const SkinModeInformation &modeInfo);

    void triggerPlaybackIntent() { emit playbackIntent(); }

signals:
    void playbackIntent();
    void currentIndexChanged();

public slots:
    void removeSelectedItems();
    void setSortedColumn(PlayListView::Columns column);

protected slots:
    virtual void selectionChanged(const QItemSelection &selected, const QItemSelection &deselected);
    virtual void currentChanged(const QModelIndex &current, const QModelIndex &previous);
    virtual void itemClicked();

protected:
    void mousePressEvent(QMouseEvent *event);

private slots:
    void graphExecutionStarted(bool &stopIt);
    void updateView();
    void playbackControllerActiveItemChanged(int newActiveItemIndex, int oldActiveItemIndex);
    void playbackControllerPlaybackStatusChanged();
    void updateSelectedPlayListItem();

private:
    friend class PlayListItemDelegate;
    PlayListAspectListViewRenderer *aspectSource_;
    PlaybackController *playbackController_;
    PlayListItemDelegate delegate_;

    int updateCount_;

    bool autoScaleColumnsEnabled_;
    //QValueList<float> headerSectionSizes_;
    bool userIsResizingSection_;
    bool autoResizingSections_;

    QPalette evenRowColorGroup_;
    QPalette oddRowColorGroup_;
    const QPixmap *playIndicatorImage_;
    const QPixmap *pauseIndicatorImage_;
    const QPixmap *errorIndicatorImage_;

    InputMode inputMode_;

    bool itemOrderDirty_;

    bool buttonDown_;
    QPoint buttonDownPos_;
    bool moveTriggered_;

    QModelIndex activeItem_;

    QMenu *popupMenu_;

    void ensurePlayListOrderIsValid();
};

#endif	// PLAYLIST_H_INCLUDED
